#include "Burden.h"
#include "Constant.h"
#include "math.h"
#include "MathStats.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include "qfc.h"

String Burden::function = "";
String Burden::genename = "";
String Burden::genename_f = "";
String Burden::vcfInput = ""; //input annotated vcf file name here
String Burden::outputFile = "";
double Burden::mafMin = 0.0;
double Burden::mafMax = 0.5;
String Burden::chromosome = "";
bool Burden::vcfAnnotated = true;
bool Burden::CMC_binary = false;
bool Burden::CMC_counts = false;
bool Burden::MB = false;
bool Burden::burden = false;
bool Burden::LinScoreAssoc = false;
bool Burden::SCORE=true;
bool Burden::LRT = false;
bool Burden::WALD = false;
bool Burden::SKAT = false;

Burden::Burden(){}

Burden::~Burden()
{}

bool Burden::GetGenotypeVectorDosage(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum) 
{
   bool status = false;

   //mac is the total number of alternative alleles included
   mac = 0.0;
   //cum_maf is the cumulative maf in this gene
   cum_maf=0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;
   int numSamples = header.getNumSamples();

   for(int m=0;m<group.chrList[groupNum].Length();m++)
   {
      reader.set1BasedReadSection(group.chrList[groupNum][m].c_str(),group.posList[groupNum][m],group.posList[groupNum][m]+1);
      while(reader.readRecord(record))
      {
	 int n=0,nf=0;
	 double fmaf = 0.0, maf=0.0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

	 //Loop througth all samples to calculate maf
	 for(int s=0;s<numSamples;s++)
	 {
	    const std::string * geno = genoInfo.getString("DS",s);
	    const char * sample = header.getSampleName(s);

	    if(geno==NULL)
	    {
	       error("Error: Can not find DS field for sample %d ...",s);
	    }
	    if(*geno == ".")
	    {
	       continue;
	    }
	    float dose = atof((*geno).c_str());

	    //judge if this is a founder
	    int founder = trans.foundersHash.Integer(sample);
	    if(founder != -1)
	    {
	       fmaf += dose;
	       nf++;
	    }
	    maf += dose;
	    n++;
	 }

	 if(n!=0)
	    maf /= 2.0*n;
	 else 
	    continue;

	 if(nf!=0)
	    fmaf /= 2.0*nf;
	 else 
	    fmaf=maf;
	 if(maf==0.0)
	 {
	    fmaf = 0.0;
	 }
	 if(fmaf==0.0 && maf>0)
	    fmaf = maf;

	 if(maf==0.0)
	    continue;

	 if(fmaf>0.5) fmaf = 1.0-fmaf;
	 if(maf>0.5) maf = 1.0-maf;

	 //check if maf falls into the range
	 if(SVT::FounderFreq)
	 {
	    if(fmaf>mafMax || fmaf<mafMin)
	       continue;
	 }
	 else
	 {
	    if(maf>mafMax || maf<mafMin)
	       continue;
	 }
	 numVarPerGene++;
	 if(SKAT)
	    genotypeSKAT.GrowTo(trans.persons,numVarPerGene,0.0);
	 if(SVT::FounderFreq)
	    cum_maf+=fmaf;
	 else 
	    cum_maf+=maf;

	 //update the genotype vector
	 int idx=0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int j=0;j<trans.pPheno[f].Length();j++)
	    {
	       String sample = ped[trans.pPheno[f][j]].pid.c_str();

	       //get the sample number by sample ID from hash
	       int s = trans.sampleVCFIDHash.Integer(sample);
	       if(s!=-1)
	       {
		  const std::string * geno = genoInfo.getString("DS",s);
		  if(strcmp((*geno).c_str(),".")>0)
		  {
		     float dose = atof((*geno).c_str());
		     if(dose>0)
		     {
			status=true;
			if(CMC_binary && genotypeCMC_b[idx]==0)
			{
			   genotypeCMC_b[idx] = 1;
			}                    
			if(CMC_counts)
			{
			   genotypeCMC_c[idx] +=dose;
			}
			if(MB)
			{
			   if(SVT::FounderFreq)
			   {
			      genotypeMB[idx] += dose/sqrt(fmaf*(1.0-fmaf));
			   }
			   else
			   {
			      genotypeMB[idx] += dose/sqrt(maf*(1.0-maf));
			   }
			}
			if(SKAT)
			   genotypeSKAT[idx][numVarPerGene-1] = dose;
			mac += dose;
		     }
		  }
	       }
	       idx++;
	    }
	 }
      }
   }
   return status;
}

bool Burden::GetGenotypeVectorVCF(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;
   //mac is the total number of alternative alleles included
   mac = 0.0;
   //cum_maf is the cumulative maf in this gene
   cum_maf=0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;
   int numSamples = header.getNumSamples();

   for(int m=0;m<group.chrList[groupNum].Length();m++)
   {
      reader.set1BasedReadSection(group.chrList[groupNum][m].c_str(),group.posList[groupNum][m],group.posList[groupNum][m]+1);
      while(reader.readRecord(record))
      {
	 //Get allele frequencies first
	 double fmaf=0.0,maf=0.0;
	 std::vector<int> fcounts;
	 std::vector<int> counts;

	 for(int i = 0; i < numSamples; i++)
	 {
	    const char * sample = header.getSampleName(i);
	    int numGTs = record.getNumGTs(i);
	    Vector allele;
	    for(int j = 0; j < numGTs; j++)
	    {
	       allele.Push(record.getGT(i,j));
	    }
	    int bad =0;
	    for(int m=0;m<allele.Length();m++)
	    {
	       bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	    }
	    if(bad>0)
	    {
	       //there is at least one allele missing, so skip this sample
	       continue;
	    }
	    //judge if this is a founder
	    int founder = trans.foundersHash.Integer(sample);

	    for(int m=0;m<allele.Length();m++)
	    {
	       if(allele[m] >= counts.size())
	       {
		  counts.resize(allele[m]+1);
	       }
	       ++counts[allele[m]];
	       if(founder != -1)
	       {
		  if(allele[m] >= fcounts.size())
		  {
		     fcounts.resize(allele[m]+1);
		  }
		  ++fcounts[allele[m]];
	       }
	    }
	    //Only consider diallelic markers
	    if(counts.size()>2)
	    {
	       fmaf = -1.0;
	       maf = -1.0;
	       continue;
	    }
	 }

	 //if mornomorphic, then skip this marker
	 if(counts.size()==1 || fcounts.size()==1)
	 {
	    fmaf = 0.0;
	    maf = 0.0;
	 }
	 else
	 {
	    if(fmaf!=-1.0)
	    {
	       double sum = 0.0;
	       for(int i=0;i<counts.size();i++)
	       {
		  sum+= counts[i];
	       }
	       maf = counts[1]/sum;
	       sum = 0.0;
	       for(int i=0;i<fcounts.size();i++)
	       {
		  sum+= fcounts[i];
	       }
	       if(sum>0)
		  fmaf = fcounts[1]/sum;

	       if(fmaf==0.0 && maf>0)
		  fmaf = maf; 
	    }
	 }
	 //Fill in genotype if SNP is qualified
	 if(maf==0.0)
	 {
	    fmaf = 0.0;
	 }
	 if(fmaf==0 && maf>0)
	    fmaf = maf;

	 if(SVT::FounderFreq)
	 {
	    maf = fmaf;
	 }
	 if(maf<=0.0)
	    continue;

	 if(maf>0.5) maf = 1.0-maf;

	 //check if maf falls into the range
	 if(maf>mafMax || maf<mafMin)
	    continue;
	 numVarPerGene++;
	 cum_maf+=maf;
	 if(SKAT)
	    genotypeSKAT.GrowTo(trans.persons,numVarPerGene,0.0);

	 //update genotype vectors 
	 //Loop through samples according to trans.pPheno 
	 //read in dosage into genotype vector
	 int idx=0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int j=0;j<trans.pPheno[f].Length();j++)
	    {
	       String sample = ped[trans.pPheno[f][j]].pid.c_str();

	       //get the sample number by sample ID from hash
	       int s = trans.sampleVCFIDHash.Integer(sample);
	       if(s!=-1)
	       {
		  int numGTs = record.getNumGTs(s);
		  Vector allele;

		  for(int j = 0; j < numGTs; j++)
		  {
		     allele.Push(record.getGT(s,j));
		  }
		  int bad =0;
		  for(int m=0;m<allele.Length();m++)
		  {
		     bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
		  }
		  //if any allele is missing then use 2*maf to impute
		  if(bad==0)
		  {
		     int sum=0;
		     for(int i=0;i<allele.Length();i++)
			sum+=allele[i];
		     if(sum>0)
		     {
			status = true;
			if(MB)
			   genotypeMB[idx] += sum/sqrt(maf*(1.0-maf));
			if(CMC_binary)
			{
			   if(sum>0||genotypeCMC_b[idx]==0)
			      genotypeCMC_b[idx] += 1;
			}
			if(CMC_counts)
			   genotypeCMC_c[idx] += sum;
			if(SKAT)
			{
			   genotypeSKAT[idx][numVarPerGene-1] = sum; 
			}
			mac += sum;
		     }
		  }
	       }
	       idx++;
	    }
	 }
      }
   }
   return status;
}

bool Burden::GetGenotypeVectorPED(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;
   //mac is the total number of alternative alleles included
   mac = 0.0;
   //cum_maf is the cumulative maf in this gene
   cum_maf=0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;
   StringArray SNPNameList;
   for(int m=0;m<group.chrList[groupNum].Length();m++)
   {
      String markerName = chrList[groupNum][m];
      markerName += ":";
      markerName += posList[groupNum][m];
      SNPNameList.Push(markerName);
   }
   //Now filter by maf calculated from pedigree file
   Vector freq; 
   int minor,markerid;  
   IntArray key,minorAllele;
   int length = SNPNameList.Length();
   SNPcount = 0;
   for(int m=0;m<length;m++)
   {
      markerid = ped.markerLookup.Integer(SNPNameList[m]);
      //markerid = ped.GetMarkerID(SNPNameList[m]);
      if(markerid==-1)
      {
	 continue;
      }
      bool tmpF = EstimateFrequencies(ped,markerid,FREQ_ALL);
      freq.Copy(ped.GetMarkerInfo(markerid)->freq);
      /*
	 for(int i=0;i<freq.Length();i++)
	 {
	 printf("%g ",freq[i]);
	 }
	 printf("\n");
       */
      markerMaf = 0.0;
      if(freq.Length()==3)
      {
	 if(freq[1] < freq[2]) { minor=1; markerMaf = freq[1]; }
	 else { minor=2; markerMaf = freq[2];}
	 if(markerMaf> mafMin && markerMaf <= mafMax)
	 {
	    key.Push(markerid);
	    minorAllele.Push(minor);
	    mafArray.Push(markerMaf); 
	    SNPcount++;
	    //  printf("maf is:%g;marker is: %d\n",markerMaf,markerid);
	    // printf("marker name is:%s\n",SNPNameList[m].c_str());
	 }
      } //Now qualified SNPs are in key.
      //printf("freq.Length %d, markerMaf %g\n",freq.Length(),markerMaf);
   }
   /*
      printf("SNPNameList is:\n");
      for(int i=0;i<SNPNameList.Length();i++)
      printf("%s\t",SNPNameList[i].c_str());
      printf("\n");

      printf("Key is:\n");
      for(int i=0;i<key.Length();i++)
      printf("%d\t",key[i]);
      printf("\n");
      printf("mafs are: \n");
      for(int i=0;i<key.Length();i++)
      printf("%g\t",mafArray[i]);
      printf("\n");
      printf("key length is: %d\n",key.Length());
    */
   bool XStatus = false;

   if(key.Length()==0) goto skip_to_point;

   if(AutoFit::fitX)
   {
      String markerName = ped.markerNames[markerid];
      int found = ped.markerNames[markerid].FastFindChar('X',0);
      StringArray tmp;
      tmp.AddTokens(markerName,":");
      if(found!=-1)
      {
	 if(atoi(tmp[1].c_str()) > SVT::Xstart && atoi(tmp[1].c_str())< SVT::Xend)
	 {
	    XStatus=true;
	 }
      }
   }

   if(SKAT)
   {
      int idx=0;
      genotypeSKAT.Dimension(trans.persons,key.Length());
      genotypeSKAT.Zero();

      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
	    for(int m=0; m<key.Length(); m++)
	    {
	       if(!ped[trans.pPheno[f][p]].isGenotyped(key[m]))
	       {
		  genotypeSKAT[idx][m] = 2.0*mafArray[m];
	       }
	       else if(ped[trans.pPheno[f][p]].markers[key[m]].isHeterozygous())
	       {
		  if(XStatus)
		  {
		     if(ped[trans.pPheno[f][p]].sex==2)
		     {
			genotypeSKAT[idx][m] = 1.0;
			status=true;
		     }
		     if(ped[trans.pPheno[f][p]].sex==1)
		     {
			genotypeSKAT[idx][m] = 2.0;
			status=true;
		     }
		  }
		  else
		  {
		     genotypeSKAT[idx][m] = 1.0;
		     status=true;
		  }
	       } else if((ped[trans.pPheno[f][p]].markers[key[m]].isHomozygous() && ped[trans.pPheno[f][p]].markers[key[m]].hasAllele(minorAllele[m]))) 
	       {
		  genotypeSKAT[idx][m] = 2.0;
		  status=true;
	       }
	    }
	    idx++;
	 }
      }
      for(int i=0;i<genotypeSKAT.cols;i++)
      {
	 double mean = 0.0;
	 for(int j=0; j<genotypeSKAT.rows;j++)
	 {
	    mean += genotypeSKAT[j][i];
	 }
	 mean /= genotypeSKAT.rows;
	 for(int j=0;j<genotypeSKAT.rows;j++)
	 {
	    genotypeSKAT[j][i] -= mean;
	 } 
      } 
   }

   if(MB) 
   {
      int idx=0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++) 
	 {
	    for(int m=0; m<key.Length(); m++)
	    {
	       if(!ped[trans.pPheno[f][p]].isGenotyped(key[m]))
	       {
		  genotypeMB[idx] += 2.0*mafArray[m]/sqrt(mafArray[m]*(1.0-mafArray[m]));
		  //genotype[idx]=2.0*markerMaf;        
	       }
	       else if(ped[trans.pPheno[f][p]].markers[key[m]].isHeterozygous())
	       {
		  if(XStatus)
		  {
		     if(ped[trans.pPheno[f][p]].sex==2)
		     {
			genotypeMB[idx] += 1.0/sqrt(mafArray[m]*(1.0-mafArray[m]));
			status=true;
		     }
		     if(ped[trans.pPheno[f][p]].sex==1)
		     {
			genotypeMB[idx] += 2.0/sqrt(mafArray[m]*(1.0-mafArray[m]));
			status=true;
		     }
		  }
		  else
		  {
		     genotypeMB[idx] += 1.0/sqrt(mafArray[m]*(1.0-mafArray[m]));
		     status=true;
		  }
	       } else if((ped[trans.pPheno[f][p]].markers[key[m]].isHomozygous() && ped[trans.pPheno[f][p]].markers[key[m]].hasAllele(minorAllele[m]))) {
		  //genotypeMB[idx] += 2.0/sqrt(trans.persons*mafArray[m]*(1.0-mafArray[m]));
		  genotypeMB[idx] += 2.0/sqrt(mafArray[m]*(1.0-mafArray[m]));
		  status=true;
	       } 
	    }
	    idx++;
	 }
      }
      double mean = genotypeMB.Average();
      genotypeMB.Add(-mean);
   }

   if(CMC_binary) 
   {
      int idx=0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++) {
	    for(int m=0; m<key.Length(); m++)
	    {   
	       if(ped[trans.pPheno[f][p]].markers[key[m]].isHeterozygous() || (ped[trans.pPheno[f][p]].markers[key[m]].isHomozygous() && ped[trans.pPheno[f][p]].markers[key[m]].hasAllele(minorAllele[m]))) 
	       {
		  genotypeCMC_b[idx]=1;
		  status=true;
		  break;
	       }
	    }
	    idx++;
	 }
      }
      double mean = genotypeCMC_b.Average();
      genotypeCMC_b.Add(-mean);
   } 

   if(CMC_counts) 
   {
      int idx=0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++) 
	 {
	    for(int m=0; m<key.Length(); m++)
	    {   
	       if(!ped[trans.pPheno[f][p]].isGenotyped(key[m]))
	       {
		  genotypeCMC_c[idx] += 2.0*mafArray[m];
		  //genotype[idx]=2.0*markerMaf;        
	       }
	       else if(ped[trans.pPheno[f][p]].markers[key[m]].isHeterozygous())
	       { 
		  if(XStatus)
		  {
		     if(ped[trans.pPheno[f][p]].sex==2)
		     {
			genotypeCMC_c[idx] += 1;
			status=true;
		     }
		     if(ped[trans.pPheno[f][p]].sex==1)
		     {
			genotypeCMC_c[idx] += 2;
			status=true;
		     }
		  }
		  else
		  {
		     genotypeCMC_c[idx] += 1;
		     status=true;
		  }
	       } 
	       else if(ped[trans.pPheno[f][p]].markers[key[m]].isHomozygous() && ped[trans.pPheno[f][p]].markers[key[m]].hasAllele(minorAllele[m])) 
	       {
		  genotypeCMC_c[idx] += 2;
		  status=true;
	       }
	    }
	    idx++;
	 }
      }
      double mean = genotypeCMC_c.Average();
      genotypeCMC_c.Add(-mean);
   }
skip_to_point:
   return status;
}

bool Burden::GetGenotypeVector(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status;
   if(MB)
   {
      genotypeMB.Dimension(trans.persons);
      genotypeMB.Zero();
   }
   if(CMC_binary)
   {
      genotypeCMC_b.Dimension(trans.persons);
      genotypeCMC_b.Zero();
   }
   if(CMC_counts)
   {
      genotypeCMC_c.Dimension(trans.persons);
      genotypeCMC_c.Zero();
   }
   if(SKAT)
      genotypeSKAT.Dimension(trans.persons,0);

   printf("Now working on group %s ...\n",annoGroups[groupNum].c_str());

   if(SVT::genoFromPed)
      status = GetGenotypeVectorPed(trans,ped,group,groupNum);
   if(SVT::genoFromVCF)
      status = GetGenotypeVectorVCF(trans,ped,group,groupNum);
   if(SVT::dosage)
      status = GetGenotypeVectorDosage(trans,ped,group,groupNum);

   return status;
}

void Burden::Calculate(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group)
{
   int traitNum = engine.GetTraitID(ped,traitName.c_str());
   FILE * MBScoreOut;
   FILE * CMC_bScoreOut;
   FILE * CMC_cScoreOut;
   FILE * MBLRTOut;
   FILE * CMC_bLRTOut;
   FILE * CMC_cLRTOut;
   FILE * MBWALDOut;
   FILE * CMC_bWALDOut;
   FILE * CMC_cWALDOut;
   FILE * MBLinScoreOut;
   FILE * CMC_bLinScoreOut;
   FILE * CMC_cLinScoreOut;
   FILE * SKATOut;

   if(SCORE)
   {
      if(MB) 
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.score.txt";
	 MBScoreOut = fopen(MBfilename,"wt");
	 PrintHeader(MBScoreOut ,trans,engine,traitName);
      }

      if(CMC_binary) 
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.score.txt";
	 CMC_bScoreOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bScoreOut ,trans,engine,traitName);
      }

      if(CMC_counts) 
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.score.txt";
	 CMC_cScoreOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cScoreOut,trans,engine,traitName);
      }
   }

   if(WALD)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.wald.txt";
	 MBWALDOut = fopen(MBfilename,"wt");
	 PrintHeader(MBWALDOut ,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.wald.txt";
	 CMC_bWALDOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bWALDOut ,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.wald.txt";
	 CMC_cWALDOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cWALDOut,trans,engine,traitName);
      }
   }

   if(LinScoreAssoc)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.LinScore.txt";
	 MBLinScoreOut = fopen(MBfilename,"wt");
	 PrintHeader(MBLinScoreOut ,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.LinScore.txt";
	 CMC_bLinScoreOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bLinScoreOut ,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.LinScore.txt";
	 CMC_cLinScoreOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cLinScoreOut,trans,engine,traitName);
      }
   }

   if(LRT)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.LRT.txt";
	 MBLRTOut = fopen(MBfilename,"wt");
	 PrintHeader(MBLRTOut,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.LRT.txt";
	 CMC_bLRTOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bLRTOut,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.LRT.txt";
	 CMC_cLRTOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cLRTOut,trans,engine,traitName);
      }
   }

   if(SKAT)
   {
      String SKATfilename;
      SKATfilename = outputFile + traitName + ".SKAT.txt";
      SKATOut = fopen(SKATfilename,"wt");
      PrintSKATHeader(SKATOut,trans,engine,traitName);
   }

   double L0,L1;

   if(LRT)
   {
      if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
      {
	 L0=engine.EvaluateMulti(trans,false);
      }
      else
      {
	 L0 = engine.Evaluate(engine.deltaHat,trans,false);
	 printf("L0 is:%g\n",L0);
      }
   }

   int bad =0;
   for(int i=0;i<group.annoGroups.Length();i++)
   {
      String geneName = group.annoGroups[i];
      if(GetGenotypeVector(trans,ped,group,i))
      {
	 Vector transGenoMB;
	 Vector transGenoCMC_b;
	 Vector transGenoCMC_c;

	 if(MB)
	    transGenoMB.Product(trans.transU,genotypeMB);
	 if(CMC_binary)
	    transGenoCMC_b.Product(trans.transU,genotypeCMC_b);
	 if(CMC_counts)
	    transGenoCMC_c.Product(trans.transU,genotypeCMC_c);

	 if(LRT || WALD)
	 {
	    FastTransform  fullTrans;
	    FastFit fullEngine;
	    fullTrans.persons = trans.persons;
	    fullTrans.fixedEffects = trans.fixedEffects +1;

	    if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	    {
	       int covariateId = ped.GetCovariateID("MarkerTested");

	       fullTrans.pPheno = new IntArray [ped.familyCount];
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  fullTrans.pPheno[f].Dimension(0);
		  for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
		  {
		     if(ped[i].isPhenotyped(traitNum) && (!engine.useCovariates || ped[i].isFullyControlled()))
		     {
			fullTrans.pPheno[f].Push(i);
		     }
		  }
	       }
	       int index =0;
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  if (fullTrans.pPheno[f].Length() > 0)
		  {
		     int count = fullTrans.pPheno[f].Length();
		     for(int i=0;i<count;i++)
		     {
			if(MB)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeMB[index];
			if(CMC_binary)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeCMC_b[index];
			if(CMC_counts)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeCMC_c[index];
			index++;
		     }
		  }
	       }
	       printf("ped.covariateCount is%d\n",ped.covariateCount);

	       AutoFit fullAutoEngine;
	       fullAutoEngine.useCovariates = true;
	       fullAutoEngine.fitSharedEnvironment = AutoFit::fitSharedEnvironment;
	       fullAutoEngine.fitX = AutoFit::fitX;

	       //Fit the alternative
	       fullAutoEngine.FitPolygenicModels(ped,traitNum,true);

	       fullTrans.Y.Copy(trans.Y);
	       fullTrans.X.Copy(trans.X);
	       fullTrans.X.GrowTo(fullTrans.persons,trans.X.cols+1,0.0);
	       fullTrans.UY.Dimension(fullTrans.persons);
	       fullTrans.UX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.transU.Dimension(0,0);
	       fullTrans.D.Dimension(0);

	       if(MB)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeMB[i];
	       if(CMC_binary)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeCMC_b[i];
	       if(CMC_counts)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeCMC_c[i];

	       fullTrans.traitVar = fullAutoEngine.variance;
	       fullTrans.TransformPedkinSibEnv(fullAutoEngine, ped);
	       fullEngine.betaHat.Copy(fullAutoEngine.betaHat);
	    }
	    else 
	    {
	       fullTrans.D.Copy(trans.D);
	       fullTrans.UY.Copy(trans.UY);
	       fullTrans.transU.Copy(trans.transU);
	       fullTrans.UD.Dimension(fullTrans.persons,fullTrans.persons);

	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.UX.Copy(trans.UX);
	       fullTrans.UX.GrowTo(trans.UX.rows,trans.UX.cols+1,0.0);
	    }

	    String method;
	    if(LRT)
	    {
	       if(MB)
	       {
		  method = "MB";
		  MerlinLRT(L0,fullTrans,MBLRTOut,transGenoMB,geneName,fullEngine,method);
	       }
	       if(CMC_binary)
	       {
		  method = "CMC_binary";
		  MerlinLRT(L0,fullTrans,CMC_bLRTOut,transGenoCMC_b,geneName,fullEngine,method);
	       }
	       if(CMC_counts)
	       {
		  method = "CMC_counts";
		  MerlinLRT(L0,fullTrans,CMC_cLRTOut,transGenoCMC_c,geneName,fullEngine,method);
	       }
	    }

	    if(WALD)
	    {
	       if(MB)
	       {
		  method = "MB";
		  WaldAssoc(fullTrans,MBWALDOut,transGenoMB,geneName,fullEngine,method);
	       }
	       if(CMC_binary)
	       {
		  method = "CMC_binary";	
		  WaldAssoc(fullTrans,CMC_bWALDOut,transGenoCMC_b,geneName,fullEngine,method);
	       }
	       if(CMC_counts)
	       {
		  method = "CMC_counts";	
		  WaldAssoc(fullTrans,CMC_cWALDOut,transGenoCMC_c,geneName,fullEngine,method);
	       }
	    }
	 }

	 String method;
	 if(LinScoreAssoc)
	 {
	    if(MB)
	    {
	       method = "MB";
	       LinAssoc(trans,engine,MBLinScoreOut,genotypeMB,geneName,method);
	    }
	    if(CMC_binary)
	    {
	       method = "CMC_binary";
	       LinAssoc(trans,engine,CMC_bLinScoreOut,genotypeCMC_b,geneName,method);
	    }
	    if(CMC_counts)
	    {
	       method = "CMC_counts";
	       LinAssoc(trans,engine,CMC_cLinScoreOut,genotypeCMC_c,geneName,method);
	    }
	 }

	 if(SCORE)
	 {
	    /*there are Merlin type score test */
	    if(MB)
	    {
	       method = "MB";
	       MerlinFastAssoc(trans,MBScoreOut,transGenoMB,geneName,engine,method);
	    }
	    if(CMC_binary)
	    {
	       method = "CMC_binary";
	       MerlinFastAssoc(trans,CMC_bScoreOut,transGenoCMC_b,geneName,engine,method);
	    }
	    if(CMC_counts)
	    {
	       method = "CMC_counts";
	       MerlinFastAssoc(trans,CMC_cScoreOut,transGenoCMC_c,geneName,engine,method);
	    }
	 }
	 if(SKAT)
	 {
	    SKATScoreAssoc(trans,SKATOut,geneName,engine);
	 }
      } 
      else 
      {
	 bad++;
	 if(SCORE)
	 {
	    if(MB)
	    {
	       fprintf(MBScoreOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       fprintf(CMC_bScoreOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       fprintf(CMC_cScoreOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(LinScoreAssoc)
	 {
	    if(MB)
	    {
	       fprintf(MBLinScoreOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       fprintf(CMC_bLinScoreOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       fprintf(CMC_cLinScoreOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(LRT)
	 {
	    if(MB)
	    {
	       fprintf(MBLRTOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       fprintf(CMC_bLRTOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       fprintf(CMC_cLRTOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(WALD)
	 {
	    if(MB)
	    {
	       fprintf(MBWALDOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_binary)
	    {
	       fprintf(CMC_bWALDOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	    if(CMC_counts)
	    {
	       fprintf(CMC_cWALDOut,"%s\t-\t-\t-\t-\t-\n",geneName.c_str());
	    }
	 }
	 if(SKAT)
	 {
	    fprintf(SKATOut,"%s\t-\t-\n",geneName.c_str());
	 }
      }
   }
   if(SCORE)
   {
      if(MB)
	 PrintFooter(MBScoreOut,bad);
      if(CMC_binary)
	 PrintFooter(CMC_bScoreOut,bad);
      if(CMC_counts)
	 PrintFooter(CMC_cScoreOut,bad);
   }
   if(LinScoreAssoc)
   {
      if(MB)
	 PrintFooter(MBLinScoreOut,bad);
      if(CMC_binary)
	 PrintFooter(CMC_bLinScoreOut,bad);
      if(CMC_counts)
	 PrintFooter(CMC_cLinScoreOut,bad);
   }
   if(LRT)
   {
      if(MB)
	 PrintFooter(MBLRTOut,bad);
      if(CMC_binary)
	 PrintFooter(CMC_bLRTOut,bad);
      if(CMC_counts)
	 PrintFooter(CMC_cLRTOut,bad);
   }
   if(WALD)
   {
      if(MB)
	 PrintFooter(MBWALDOut,bad);
      if(CMC_binary)
	 PrintFooter(CMC_bWALDOut,bad);
      if(CMC_counts)
	 PrintFooter(CMC_cWALDOut,bad);
   }
   if(SKAT)
      PrintFooter(SKATOut,bad);

   if(MB)
   {
      if(SCORE) fclose(MBScoreOut);
      if(LinScoreAssoc) fclose(MBLinScoreOut);
      if(LRT) fclose(MBLRTOut);
      if(WALD) fclose(MBWALDOut);
   }
   if(CMC_binary)
   {
      if(SCORE) fclose(CMC_bScoreOut);
      if(LinScoreAssoc) fclose(CMC_bLinScoreOut);
      if(LRT) fclose(CMC_bLRTOut);
      if(WALD) fclose(CMC_bWALDOut);
   } 
   if(CMC_counts)
   {
      if(SCORE) fclose(CMC_cScoreOut);
      if(LinScoreAssoc) fclose(CMC_cLinScoreOut);
      if(LRT) fclose(CMC_cLRTOut);
      if(WALD) fclose(CMC_cWALDOut);
   }
   if(SKAT)
      fclose(SKATOut);
   //Restoring the pedigree covariate settings.
   if((AutoFit::fitX || AutoFit::fitSharedEnvironment) && (LRT || WALD))
   {
      ped.covariateNames.Delete(ped.covariateNames.Length()-1);
      ped.covariateLookup.Delete("MarkerTested");
      ped.covariateCount--;
      for(int i=0;i<ped.count;i++)
	 ped[i].covariates[ped.covariateCount] = 0.0;
   }
}

void Burden::Run(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group)
{
   if(burden || SKAT)
   {
      if(!vcfAnnotated)
      {
	 error("\nERROR! VCF file needs to be annotated in INFO field.\n");
      }

      if(SKAT || LinScoreAssoc)
      {
	 fixedEff.Dimension(trans.persons);
	 //	 beta_hat.Dimension(trans.fixedEffects);
	 //beta_hat.Copy(engine.betaHat);
	 //engine.EstimateAnalyticalBeta(trans.UDY,trans, beta_hat);
	 for(int i=0;i<trans.persons;i++)
	 {
	    fixedEff[i] = trans.UDX[i].InnerProduct(engine.betaHat);
	 }
	 residuals_original.Copy(trans.UDY);
	 residuals_original.Subtract(fixedEff);

	 sigma2_hat =1.0;
	 //sigma2_hat = residuals_original.SumSquares()/trans.persons;
	 //printf("sigma2_hat is:%g\n",sigma2_hat);
      }

      if(SKAT)
      {
	 printf("Start doing SKAT ...\n");
	 P0.Dimension(trans.persons,trans.persons);
	 Matrix tmp;
	 tmp.Product(trans.UDX,trans.inv);
	 Matrix X;
	 X.Transpose(trans.UDX);
	 P0.Product(tmp,X);
	 for(int i=0;i<trans.persons;i++)
	 {
	    P0[i][i] = 1.0 - P0[i][i];
	    //P0[i][i] *= sigma2_hat;
	    for(int j=i+1;j<trans.persons;j++)
	    {
	       P0[i][j] = -P0[i][j]; 
	       P0[j][i] = P0[i][j]; 
	    }
	 }
	 /*
	    printf("P0 is\n");
	    for(int i=0;i<trans.persons;i++)
	    {
	    for(int j=0;j< trans.persons;j++)
	    {
	    printf("%g ",P0[i][j]);
	    }
	    printf("\n");
	    }
	  */
	 printf("SKAT initialization done. \n");
      }

      reader.open(vcfInput,header);
      reader.readVcfIndex();

      if(MB ||CMC_binary||CMC_counts)
	 printf("Start doing the following Burden test: ");

      if(MB)
      {
	 printf("Madsen-Browning ");
      }

      if(CMC_binary)
      {
	 printf("CMC_binary ");
      }

      if(CMC_counts)
      {
	 printf("CMC_counts ");
      }
      printf(".\n");

      if(SKAT)
	 printf("Start doing the SKAT score test ...\n");

      Calculate(ped,trans,engine,traitName,group);
   }
}

void Burden::LinAssoc(FastTransform & trans,FastFit & engine,FILE * out, Vector & genotype,String & geneName,String & method)
{
   //Fill in UDX, UD, and UDY
   //trans.FinalizeTransform(engine.sigma_g2Hat, engine.sigma_e2Hat);
   Vector transGeno;
   transGeno.Product(trans.UD,genotype);
   double pvalue,numerator,denominator,chisq,effSize;
   Vector denominator_2;
   denominator_2.Dimension(trans.fixedEffects);
   numerator = 0.0;
   denominator = 1.0e-10;
   denominator_2.Zero();

   for(int i=0;i<trans.persons;i++)
   {
      numerator += residuals_original[i]*transGeno[i];
      denominator += transGeno[i]*transGeno[i];
      for(int j=0;j<trans.fixedEffects;j++)
      {
	 denominator_2[j] += transGeno[i]*trans.UDX[i][j];
      }
   }

   Vector tmp; //tmp is T(sum_SiZi)*inv
   tmp.Dimension(trans.fixedEffects);

   for(int i=0;i<trans.fixedEffects;i++)
   {
      tmp[i] = trans.inv[i].InnerProduct(denominator_2);
   }

   denominator -= tmp.InnerProduct(denominator_2);
   denominator *= sigma2_hat;

   /*
      Vector U;
      U.Dimension(trans.persons);
   //fill in U for variance
   for(int p=0;p<trans.persons;p++)
   {
   U[p] = residuals_original[p] *(transGeno[p] - tmp.InnerProduct(trans.UDX[p]));
   }
   double var = U.SumSquares();
    */
   if(denominator==0) chisq=0.0;
   else chisq = numerator*numerator/denominator;
   pvalue = chidist(chisq, 1);
   effSize = -numerator/denominator;
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      //effSize /= (SNPCount*1/)
      fprintf(out,"%s\t%d\t%g\t%g\t%g%s\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,eff.c_str(),pvalue);
   }
   else
   {
      fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,pvalue);
   }
}

void Burden::MerlinLRT(double L0,FastTransform & fullTrans,FILE * out, Vector & transGeno,String & geneName,FastFit & fullEngine,String & method)
{
   double L1,chisq,effSize,pvalue;
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      L1=fullEngine.EvaluateMulti(fullTrans,false);

      chisq = 2.0*(L1-L0);
      if(chisq<0)
	 chisq = -chisq;
      //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
      pvalue = chidist(chisq, 1);
      effSize = fullEngine.betaHat[fullTrans.fixedEffects-1];
   }
   else 
   {
      for(int row=0;row<fullTrans.persons;row++)
      {
	 fullTrans.UX[row][fullTrans.fixedEffects-1] =transGeno[row];
      }
      fullEngine.OptimizeDelta(0.0001,fullTrans);
      //fullEngine.deltaHat = fullEngine.min;
      fullEngine.GetBeta(fullEngine.deltaHat,fullTrans);
      L1 = fullEngine.Evaluate(fullEngine.deltaHat,fullTrans,false);

      chisq = 2.0*(L1-L0);
      //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
      if(chisq<0)
	 chisq = -chisq;
      //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
      pvalue = chidist(chisq, 1);
      effSize = fullEngine.beta[fullTrans.fixedEffects-1];
   }
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      fprintf(out,"%s\t%d\t%g\t%g\t%g%s\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,eff.c_str(),pvalue);
   }
   else
   {
      fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,pvalue);
   }
}

//Genotype here should be the transformed genotype
void Burden::WaldAssoc(FastTransform & fullTrans,FILE * out, Vector & genotype,String & geneName,FastFit & fullEngine,String & method)
{
   double pvalue,chisq,effSize;
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
      fullTrans.UDY.Dimension(fullTrans.persons);
      fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

      fullTrans.FinalizeTransform(0.0,0.0);
      effSize = fullEngine.betaHat[fullTrans.fixedEffects-1];
      chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1];
      pvalue = chidist(chisq, 1);
   }
   else
   {
      for(int row=0;row<fullTrans.persons;row++)
      {
	 fullTrans.UX[row][fullTrans.fixedEffects-1] = genotype[row];
      }
      fullEngine.OptimizeDelta(0.0001,fullTrans);
      //fullEngine.deltaHat = fullEngine.min;
      fullEngine.GetBeta(fullEngine.deltaHat,fullTrans);
      effSize = fullEngine.beta[fullTrans.fixedEffects-1];
      fullEngine.sigma_g2Hat = fullEngine.GetSigma_g(fullEngine.deltaHat,fullEngine.beta,fullTrans);
      fullEngine.sigma_e2Hat = fullEngine.sigma_g2Hat * fullEngine.deltaHat;

      fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
      fullTrans.UDY.Dimension(fullTrans.persons);
      fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

      fullTrans.FinalizeTransform(fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
      chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1];
      pvalue = chidist(chisq,1); 
   }
   //double pvalue = 2.0*tdist(t0, fullTrans.persons-fullTrans.fixedEffects);
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      //effSize /= (SNPCount*1/)
      fprintf(out,"%s\t%d\t%g\t%g\t%g%s\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,eff.c_str(),pvalue);
   }
   else
   {
      fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,pvalue);
   }
}

//TODO:does not work when shared env is fitted
void Burden::SKATScoreAssoc(FastTransform & trans,FILE * out,String & geneName,FastFit & engine)
{
   printf("Doing SKAT on gene %s.\n",geneName.c_str());
   Matrix transGeno;
   transGeno.Product(trans.UD,genotypeSKAT);
   double Qstat,pvalue;

   Matrix GWG;
   GWG.Dimension(trans.persons,trans.persons);
   GWG.Zero();
   Vector weight;
   weight.Dimension(genotypeSKAT.cols);

   for(int i=0;i<weight.Length();i++)
   {
      double alpha = 1.0;
      double beta=25.0;
      double beta_density = GetBetaDensity(alpha,beta,mafArray[i]);
      weight[i] = beta_density * beta_density;
   }

   for(int i=0;i<trans.persons;i++)
   {
      for(int m=0;m<genotypeSKAT.cols;m++)
      {
	 GWG[i][i] += weight[m]*transGeno[i][m]*transGeno[i][m];
      }
      for(int j=i+1;j<trans.persons;j++)
      {
	 for(int m=0;m<genotypeSKAT.cols;m++)
	 {
	    GWG[i][j] += transGeno[i][m]*transGeno[j][m]*weight[m];
	 }
	 GWG[j][i] = GWG[i][j];
      }
   }

   Vector tmp;
   tmp.Dimension(trans.persons);
   for(int i=0;i<tmp.Length();i++)
   {
      tmp[i] = residuals_original.InnerProduct(GWG[i]);
   }
   Qstat = tmp.InnerProduct(residuals_original);
   /*
      printf("Qstat is %g; number of variant is %d\n",Qstat,weight.Length());
      printf("weights are:\n");
      for(int i=0;i<weight.Length();i++)
      printf("%g ",weight[i]);
      printf("\n");
    */
   int n = weight.Length();
   double * lambda = new double [n];

   CalculateLambda(weight,lambda,transGeno); //now lambda is filled in
   /*
      printf("Lambdas are:\n");
      for(int i=0;i<weight.Length();i++)
      printf("%g ",lambda[i]);
      printf("\n");
    */
   double pvaule;
   Qstat = -Qstat; //this is for using Davies algorithm
   for(int i=0;i<n;i++)
      lambda[i] *= -1.0; //this is also for using Davies algorithm
   pvalue = MixChidist(lambda, n, Qstat,"Davies");

   if(pvalue==0)
      pvalue = MixChidist(lambda, n, Qstat,"Liu");
   //Adjust effSize to make sense
   fprintf(out,"%s\t%g\n",geneName.c_str(),pvalue);
   if(lambda) delete [] lambda;
}

double Burden::GetBetaDensity(double a, double b, double x)
{
   double density;
   density = exp(gammln(a+b)-gammln(a)-gammln(b)+(a-1.0)*log(x)+(b-1.0)*log(1.0-x));
   return density;
}

void Burden::CalculateLambda(Vector & weight,double lambda [],Matrix & transGeno)
{
   //Goal of this function is to fill in lambda
   //now we only take diagnol weight matrix 
   Matrix WG;
   WG.Dimension(genotypeSKAT.cols,genotypeSKAT.rows);
   for(int i=0;i<WG.rows;i++)
   {
      for(int j=0;j<WG.cols;j++)
      {
	 WG[i][j] = transGeno[j][i] * sqrt(weight[i]);	
      }
   }
   Matrix GW;
   GW.Transpose(WG);
   Matrix tmp;
   tmp.Product(WG,P0);
   Matrix final;
   final.Product(tmp,GW);
   /*
      printf("final matrix is: \n");
      for(int i=0;i<final.rows;i++)
      {
      for(int j=0;j<final.cols;j++)
      printf("%g ",final[i][j]);
      printf("\n");
      }
    */
   SVD svd;
   svd.Decompose(final);
   //Fill in lambda here
   for(int i=0;i<weight.Length();i++)
   {
      lambda[i] = svd.w[i];
   }
}

double Burden::MixChidist(double lambda [], int n, double q_obs,String method)
{
   double * nc1 = new double [n];
   int * n1 = new int [n];
   double * trace = new double [7];
   double sigma=0.0;
   int lim1 = 10000;
   double  acc = 0.0001;
   int  ifault = 0;
   double  res = 0.0;

   for(int i=0;i<n;i++)
   {
      nc1[i] = 0.0; //noncentral parameters
      n1[i] = 1; //h in R function
   }
   for(int i=0;i<7;i++)
   {
      trace[i] = 0.0;
   }

   if(method=="Davies")
   {
      //probability will be saved in res.
      qfc(lambda,nc1,n1,& n,& sigma, & q_obs,& lim1,& acc,trace,& ifault,& res);
      return res;
   }

   if(method=="Liu")
   {
      double c1 = 0.0; 
      double c2 = 0.0;
      double c3 = 0.0;
      double c4 = 0.0;
      for(int i=0;i<n;i++)
      {
	 c1 += n1[i] * lambda[i]; 
	 c2 += n1[i] * lambda[i]*lambda[i];          
	 c3 += n1[i] * lambda[i]*lambda[i]*lambda[i];          
	 c4 += n1[i] * lambda[i]*lambda[i]*lambda[i]*lambda[i];          
      }
      for(int i=0;i<n;i++)
      {
	 c1 += nc1[i] * lambda[i];
	 c2 += 2.0*nc1[i] * lambda[i]*lambda[i];
	 c3 += 3.0*nc1[i] * lambda[i]*lambda[i]*lambda[i];
	 c4 += 4.0*nc1[i] * lambda[i]*lambda[i]*lambda[i]*lambda[i];
      }
      double s1 = c3/(sqrt(c2*c2*c2));
      double s2 = c4/(c2*c2);
      double muQ = c1;
      double sigmaQ = sqrt(2.0*c2);
      double tstar = (q_obs-muQ)/sigmaQ;
      double delta,l,a,Qq;

      if(s1*s1>s2)
      {
	 //	 a = q_obs/(s1-sqrt(s1*s1-s2));
	 a = 1.0/(s1-sqrt(s1*s1-s2));
	 delta = s1*a*a*a - a*a;
	 l = a*a-2.0*delta;
      }
      else
      {
	 //a = 1.0/s1;
	 l = 1.0/s2;
	 a=sqrt(l);
	 delta = 0.0;
	 //l = c2*c2*c2/(c3*c3);
      }
      double muX = l + delta;
      double sigmaX = sqrt(2.0)*a;

      double stat = tstar*sigmaX+muX;
      printf("stat is:%g;df is:%g;delta is: %g\n",stat,l,delta);

      if(stat<0||l<0||delta<0)
	 return -1.0;

      if(delta==0.0)
	 Qq = chidist(stat,l);
      else
	 Qq = chidist(stat,l,delta);

      return Qq;
   }
}

void Burden::MerlinFastAssoc(FastTransform & trans,FILE * out, Vector & transGeno,String & geneName,FastFit & engine,String & method)
{
   double numerator = 0.0;
   double denominator = 1.0e-10;
   double fixedEffect, chisq,pvalue,effSize;
   Vector sigma2;
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      sigma2.Copy(trans.D);
   }
   else 
   {
      sigma2.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 sigma2[i] = engine.sigma_g2Hat*trans.D[i] + engine.sigma_e2Hat;
   }
   /*
      printf("engine.g2,e2,betahat:%g %g %g %g %g %g\n",engine.sigma_g2Hat,engine.sigma_e2Hat,engine.betaHat[0],engine.betaHat[1],engine.betaHat[2],engine.betaHat[3]);
      printf("sigma2 is:\n");
      for(int i=0;i<sigma2.Length();i++)
      printf("%g ",sigma2[i]);
      printf("\nnumerator and denominator: \n");
    */
   for(int i=0;i<trans.persons;i++)
   {
      fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
      numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
      denominator += transGeno[i]*transGeno[i]/sigma2[i];
      //    printf("%d %g %g %g %g %g\n",i,fixedEffect,genotype[i],transGeno[i],numerator,denominator);
   }

   //printf("numerator and denominator are: %g, %g\n",numerator,denominator);

   if(denominator==0) chisq=0.0;
   else chisq = numerator*numerator/denominator;
   pvalue = chidist(chisq, 1);
   effSize = numerator/denominator;
   //Adjust effSize to make sense
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      //effSize /= (SNPCount*1/)
      fprintf(out,"%s\t%d\t%g\t%g\t%g%s\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,eff.c_str(),pvalue);
   }
   else
   {
      fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,cum_maf,mac,effSize,pvalue);
   }
}

void Burden::PrintFooter(FILE * output,int bad)
{
   fprintf(output,"===================================================================\n");
   fprintf(output,"\n%d %s tested.\n",geneCount,geneCount>1?"genes were":"gene was");
   fprintf(output,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");
}

void Burden::PrintHeader(FILE * output,FastTransform & trans,FastFit & engine,String & traitName)
{
   double h2 = 100.0*engine.sigma_g2Hat/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv = 100.0* engine.sigma_s2Hat/trans.traitVar;

   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, Xheritability = %6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }

   fprintf(output,"===================================================================\n");
   fprintf(output,"Gene\tNumVarInGene\tSumMAF\tInformative_AC\tEffSIZE\tpvalue\n");
}
void Burden::PrintSKATHeader(FILE * output,FastTransform & trans,FastFit & engine,String & traitName)
{
   double h2 = 100.0*engine.sigma_g2Hat/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv = 100.0* engine.sigma_s2Hat/trans.traitVar;

   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, Xheritability = %6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }

   fprintf(output,"===================================================================\n");
   fprintf(output,"gene\tpvalue_Davies\tpvalue_Liu\n");
}

